library account;

import 'dart:core';
import 'package:json_annotation/json_annotation.dart';
import 'provider/authing.dart';

part 'client.g.dart';

enum Gender {
  @JsonValue("M")
  male,
  @JsonValue("F")
  female,
  @JsonValue("U")
  unknown,
}

@JsonSerializable(fieldRename: FieldRename.snake, includeIfNull: false)
class Address {
  final String? country;
  final String? province;
  final String? city;
  final String? street;

  Address([this.country, this.province, this.city, this.street]);

  factory Address.fromJson(Map<String, dynamic> json) =>
      _$AddressFromJson(json);

  Map<String, dynamic> toJson() => _$AddressToJson(this);
}

@JsonSerializable(fieldRename: FieldRename.snake, includeIfNull: false)
class Account {
  final String userId;
  final String? username;
  final String? phone;
  final String? email;

  final String? nickname;
  final String? givenName;
  final String? familyName;
  final String? middleName;
  final String? photo;
  final Gender? gender;
  final DateTime? birthdate;
  final Address? address;

  Account(this.userId,
      [this.username,
      this.phone,
      this.email,
      this.nickname,
      this.givenName,
      this.familyName,
      this.middleName,
      this.photo,
      this.gender,
      this.birthdate,
      this.address]);

  factory Account.fromJson(Map<String, dynamic> json) =>
      _$AccountFromJson(json);

  Map<String, dynamic> toJson() => _$AccountToJson(this);
}

@JsonSerializable(fieldRename: FieldRename.snake, includeIfNull: false)
class AuthInfo {
  final Account account;
  final String idToken;
  final String? accessToken;
  final String? refreshToken;

  AuthInfo(this.account, this.idToken, [this.accessToken, this.refreshToken]);

  factory AuthInfo.fromJson(Map<String, dynamic> json) =>
      _$AuthInfoFromJson(json);

  Map<String, dynamic> toJson() => _$AuthInfoToJson(this);
}

enum StatusCode implements Comparable<StatusCode> {
  @JsonValue(0)
  none(value: 0),
  @JsonValue(200)
  ok(value: 200),
  @JsonValue(400)
  badRequest(value: 400),
  @JsonValue(401)
  unauthorized(value: 401),
  @JsonValue(402)
  paymentRequired(value: 402),
  @JsonValue(403)
  forbidden(value: 403),
  @JsonValue(404)
  notFound(value: 404),
  @JsonValue(405)
  methodNotAllowed(value: 405),
  @JsonValue(407)
  oauthRequired(value: 407),
  @JsonValue(409)
  conflict(value: 409),
  @JsonValue(410)
  gone(value: 410),
  @JsonValue(429)
  tooManyRequests(value: 429),
  @JsonValue(500)
  internalServerError(value: 500),
  @JsonValue(503)
  serviceUnavailable(value: 503),
  @JsonValue(505)
  versionNotSupported(value: 505),
  @JsonValue(600)
  localSdkError(value: 600);

  const StatusCode({
    required this.value,
  });

  final int value;

  @override
  int compareTo(StatusCode other) => value - other.value;
}

@JsonSerializable(
    genericArgumentFactories: true,
    fieldRename: FieldRename.snake,
    createFactory: false,
    includeIfNull: false)
class Result<TData> {
  final StatusCode code;
  final String message;
  final TData? data;

  Result(this.code, this.message, [this.data]);

  Result.success(this.data)
      : code = StatusCode.ok,
        message = "";

  Result.failure(this.code, this.message) : data = null;

  Map<String, dynamic> toJson() => _$ResultToJson(this, (TData v) => v);

  bool isSuccess() {
    return code == StatusCode.ok;
  }
}

abstract class OddsAccountClient {
  factory OddsAccountClient.authing(
      [String userPoolId = "", String appId = ""]) {
    final authing = Authing.singleton();
    if (userPoolId.isNotEmpty && appId.isNotEmpty) {
      authing.init(userPoolId, appId);
    }
    return authing;
  }

  Future<Result<void>> claimPhoneOtp(String phone, [String? countryCode]);

  Future<Result<AuthInfo>> registerByPhoneWithOtp(
      String phone, String password, String otp);

  Future<Result<AuthInfo>> registerByUsername(String username, String password);

  Future<Result<AuthInfo>> registerByEmail(String email, String password);

  Future<Result<AuthInfo>> loginByPhoneOtp(String phone, String otp);

  Future<Result<AuthInfo>> loginByPassword(String account, String password);

  Future<Result<void>> logout();

  OidcAuthSession createOidcAuthSession();

  Future<Result<AuthInfo>> renewToken(String refreshToken);
}

abstract class OidcAuthSession {
  Future<String> getAuthPageUri();

  Future<Result<AuthInfo>> obtainAuthInfoByRedirectUri(String redirectBackUrl);
}
