import '../client.dart';
import 'package:authing_sdk/authing.dart' as authing_sdk;
import 'package:authing_sdk/client.dart' show AuthClient;
import 'package:authing_sdk/result.dart' show AuthResult;
import 'package:authing_sdk/oidc/auth_request.dart';
import 'package:authing_sdk/oidc/oidc_client.dart';

class Authing implements OddsAccountClient {
  static final Authing _authing = Authing._internal();

  factory Authing.singleton() {
    return _authing;
  }

  Authing._internal();

  void init(String userPoolId, String appId) {
    authing_sdk.Authing.init(userPoolId, appId);
  }

  @override
  Future<Result<void>> claimPhoneOtp(String phone,
      [String? countryCode]) async {
    final result = await AuthClient.sendSms(phone, countryCode);
    return result.toOddsResult();
  }

  @override
  Future<Result<AuthInfo>> registerByPhoneWithOtp(
      String phone, String password, String otp) async {
    final result = await AuthClient.registerByPhoneCode(phone, otp, password);
    return result.toOddsAuthResult();
  }

  @override
  Future<Result<AuthInfo>> registerByUsername(
      String username, String password) async {
    final result = await AuthClient.registerByUserName(username, password);
    return result.toOddsAuthResult();
  }

  @override
  Future<Result<AuthInfo>> registerByEmail(
      String email, String password) async {
    final result = await AuthClient.registerByEmail(email, password);
    return result.toOddsAuthResult();
  }

  @override
  Future<Result<AuthInfo>> loginByPhoneOtp(String phone, String otp) async {
    final result = await OIDCClient.loginByPhoneCode(phone, otp);
    return result.toOddsAuthResult();
  }

  @override
  Future<Result<AuthInfo>> loginByPassword(
      String account, String password) async {
    final result = await OIDCClient.loginByAccount(account, password);
    return result.toOddsAuthResult();
  }

  @override
  Future<Result<void>> logout() async {
    final result = await AuthClient.logout();
    return result.toOddsResult();
  }

  @override
  OidcAuthSession createOidcAuthSession() {
    return AuthingOidcAuthSession();
  }

  @override
  Future<Result<AuthInfo>> renewToken(String refreshToken) async {
    final result =
        await OIDCClient.getNewAccessTokenByRefreshToken(refreshToken);
    return result.toOddsAuthResult();
  }
}

extension Converter on AuthResult {
  StatusCode _extractStatusCode() {
    switch (code) {
      case 200:
        return StatusCode.ok;
      case 400: // 参数格式错误
      case 2003: // 注册或登录时邮箱格式不正确
      case 2007: // 应用名非法
      case 2009: // 非法的应用类型
      case 2010: // 需提供操作的应用的 ID
      case 2012: // 缺少默认用户组
      case 2013: // 非法的应用描述
      case 2014: // 搜索用户时输入格式错误
      case 2015: // 搜索用户时搜索类型非法
      case 2016: // 解密客户端密码出错
      case 2024: // 项目描述不能超过 140 个字
      case 2028: // 请提供正确的手机号或邮箱
      case 2029: // 密码长度不能少于 6 位
      case 2030: // 一次性查询的用户不能超过 80 人
      case 2032: // 注册时需要密码
      case 2101: // 请提供应用 ID
      case 2204: // 邮箱格式不正确
      case 2205: // 缺少参数：registerInClient
      case 2211: // 缺少参数：username（用户 username）
      case 2231: // 请求参数不全
      case 5025: // 创建订单失败：价格不合法
        return StatusCode.badRequest;

      case 1639: // 需要进行二次认证
      case 2000: // 账号异常，需要输入验证码
      case 2001: // 验证码验证失败(验证码错误)
      case 2006: // 密码不正确
      case 2019: // 修改密码时需要先进行验证
      case 2020: // 尚未登录，无权限访问此请求
      case 2201: // 请输入原始密码
      case 2203: // 原始密码错误
      case 2206: // 登录信息已过期, 需重新登录
      case 2207: // 登录信息有误, 需重新登录
      case 2227: // 修改邮箱必须验证原来的邮箱
      case 2228: // 修改邮箱必须验证原来的邮箱
      case 2333: // 账号或密码错误
        return StatusCode.unauthorized;

      case 2002: // 月登录限额已用完
        return StatusCode.paymentRequired;

      case 1001: // 无权限执行此项操作
      case 2005: // 用户已被锁定
      case 2018: // 用户无权限修改此项内容
      case 2031: // 应用已禁止注册用户
      case 2040: // 人头数目限制已满
      case 2202: // 修改的信息不属于当前用户
      case 2209: // 无权限删除该用户
      case 2212: // 不能删除 root 用户
      case 2222: // 当尝试创建或更新 OAuth 应用信息时，使用了保留域名
      case 3617: // 无权添加协作者
      case 3618: // 无权添加权限项目
      case 3619: // 无权查看此用户参与协作的用户池信息
      case 3621: // 无权删除协作关系
      case 3622: // 无权查看协作者列表
      case 3624: // 无权修改协作者
        return StatusCode.forbidden;

      case 2004: // 用户不存在
      case 2011: // 应用不存在
      case 2036: // 用户不属于此用户池
      case 2037: // 填写的旧邮箱与实际邮箱不匹配
      case 2038: // 填写的旧手机与实际手机不匹配
      case 2216: // 当尝试解绑第三方 OAuth 登录方式时，没有绑定过此种 OAuth 登录
      case 2219: // 当尝试解绑邮箱时，此用户没有绑定邮箱
      case 2221: // 当尝试更新 OAuth 应用时，应用不存在
      case 2224: // 用户池不存在
      case 2401: // 移动应用 sso session 不存在
      case 3623: // 协作关系不存在
      case 4212: // OIDC 应用不存在
      case 5001: // 订单不存在
      case 7348: // SAML SP 应用不存在
        return StatusCode.notFound;

      case 2225: // updateUser 接口不能直接修改邮箱，请使用 updateEmail 接口
      case 2226: // updateUser 接口不能直接修改手机号，请使用 updatePhone 接口
        return StatusCode.methodNotAllowed;

      case 2027: // 尝试用密码登录，因未设置密码，无法验证，请通过 OAuth 登录
        return StatusCode.oauthRequired;

      case 2008: // 已拥有同名应用
      case 2026: // 用户已存在，请不要重复注册
      case 2034: // 用户名已存在
      case 2035: // 手机号已绑定
      case 2039: // 域名已被占用
      case 2200: // 该邮箱已被绑定
      case 2208: // 请换一个与现有邮箱不同的邮箱吧
      case 2213: // 当尝试绑定第三方 OAuth 登录方式时，发现已绑定过此种方式
      case 2215: // 当尝试绑定第三方 OAuth 登录方式时，将要绑定的账号已被绑定过了
      case 2217: // 当尝试解绑第三方 OAuth 登录方式或邮箱时，只有一种登录方式，故不能解绑
      case 2218: // 当尝试修改密码时，没有绑定邮箱，不允许修改
      case 2220: // 当尝试创建或更新 OAuth 应用时，已经存在此应用名
      case 2223: // 当尝试创建 OAuth 应用信息时，使用了已经被使用的域名
      case 2229: // 新邮箱和旧邮箱一样
      case 2230: // 新手机号和旧手机号一样
      case 3620: // 协作者已存在
      case 3829: // 此二级域名已被占用
        return StatusCode.conflict;

      case 2023: // 用户邮箱验证失败，原因：验证链接已过期，需重新发送
      case 2300: // 验证码过期
        return StatusCode.gone;

      case 2100: // 注册过于频繁，请稍候再试
        return StatusCode.tooManyRequests;

      case 2017: // 解析邮件模本的 meta_data(宏)命令错误
      case 2021: // 邮件发送失败，原因：无法获取邮件模版
      case 2022: // 用户邮箱验证失败，原因：无法获取邮件模板
      case 2025: // 使用默认邮件服务商出错
      case 2210: // 执行了错误的删除操作，可能原因是意图删除不存在的用户，或删除过程中出现了其它错误
      case 2214: // 读取已绑定的 OAuth 登录方式失败
      case 3012: // 宏命令执行错误
      case 3013: // 发送邮件错误，未知错误
      case 3014: // 邮件发送失败，原因：无法获取 transporter
      case 5000: // 获取订单对应应用失败
      case 5022: // 创建订单失败
      case 5023: // 创建支付宝订单失败
      case 5024: // 创建订单失败，未知错误
      case 8128: // 返回 saml assertion 时发生错误
        return StatusCode.internalServerError;

      case 1000: // 系统繁忙，请稍后再试
        return StatusCode.serviceUnavailable;

      default:
        return StatusCode.none;
    }
  }

  String _extractMessage() {
    return code == 200 ? message : "$message ($code)";
  }

  Gender _parseGender(String value) {
    switch (value) {
      case "M":
        return Gender.male;
      case "F":
        return Gender.female;
      default:
        return Gender.unknown;
    }
  }

  DateTime? _parseDateTime(String? value) {
    if (value == null || value == "null") {
      return null;
    }

    return DateTime.parse(value);
  }

  Account? _extractAccount() {
    if (user == null) {
      return null;
    }

    return Account(
      user!.id,
      user!.username,
      user!.phone,
      user!.email,
      user!.nickname,
      user!.givenName,
      user!.familyName,
      user!.middleName,
      user!.photo,
      _parseGender(user!.gender),
      _parseDateTime(user!.birthdate),
      Address(user!.country, user!.province, user!.city, user!.streetAddress),
    );
  }

  bool _isSuccess() {
    return code == 200;
  }

  Result<AuthInfo> toOddsAuthResult() {
    if (_isSuccess()) {
      return Result.success(AuthInfo(_extractAccount()!, user!.token,
          user!.accessToken, user!.refreshToken));
    }
    return Result.failure(_extractStatusCode(), _extractMessage());
  }

  Result<void> toOddsResult() {
    if (_isSuccess()) {
      return Result.success(null);
    }

    return Result.failure(_extractStatusCode(), _extractMessage());
  }
}

class AuthingOidcAuthSession implements OidcAuthSession {
  final AuthRequest _authRequest;

  AuthingOidcAuthSession() : _authRequest = AuthRequest() {
    _authRequest.createAuthRequest();
  }

  @override
  Future<String> getAuthPageUri() async {
    return await OIDCClient.buildAuthorizeUrl(_authRequest);
  }

  @override
  Future<Result<AuthInfo>> obtainAuthInfoByRedirectUri(
      String redirectBackUri) async {
    late Uri uri;
    try {
      uri = Uri.parse(redirectBackUri);
    } catch (e) {
      return Result.failure(StatusCode.localSdkError, "${e.toString()} (6000)");
    }

    final String? authCode = uri.queryParameters["code"];
    if (authCode == null) {
      return Result.failure(
          StatusCode.localSdkError, "The code parameter is missing (6001)");
    }
    final result = await OIDCClient.authByCode(authCode, _authRequest);
    return result.toOddsAuthResult();
  }
}
